(define emit (lambda args (apply simple-format #t args)
                          (newline)))

(define fixnum-shift 2)
(define fixnum-mask 3)

(define ptr-mask 7) ; mask for pointer type tag
(define ptr-mask-inv #xfffffff8) ; mask for pointer value

(define pair-tag 1)
(define vec-tag 2)
(define str-tag 3)
(define sym-tag 5)
(define closure-tag 6)

(define char-mask 255) ; character type mask
(define char-shift 8)
(define char-tag 7)

(define bool-mask 255)
(define bool-shift 8)
(define bool-tag 15)

(define wordsize 4)

(define (immediate-rep x)
  (cond ((integer? x) (logand (ash x fixnum-shift) #xffffffff))
        ((char? x) (logior (ash (char->integer x) char-shift) char-tag))
        ((boolean? x)
         (if x
             (logior (ash 1 bool-shift) bool-tag)
             bool-tag))
        ((null? x) pair-tag)
    ))

; Check whether the passed form is a primitive call (primcall) form
(define (primitive-call? form) (eq? 'primcall (car form)))

; Get the primitive operation from a passed primcall form
(define (primitive-op form) (cadr form))

; Get the Nth argument of a passed primcall form
(define (primitive-op-arg1 form) (caddr form))
(define (primitive-op-arg2 form) (cadddr form))
(define (primitive-op-arg3 form) (car (cddddr form)))

; Get all arguments of a passed primcall form
(define (primitive-op-args form) (cddr form))

(define (immediate? x) (or (integer? x) (char? x) (boolean? x) (null? x)))

(define (emit-is-eax-equal-to val)
  (emit "cmpl $~a, %eax" val) ; check eax against val
  (emit "movl $0, %eax")      ; zero eax, leaving equal flag in place
  (emit "sete %al")           ; set low bit of eax if they were equal
  (emit "sall $~a, %eax" bool-shift) ; shift the bit up to the bool position
  (emit "orl $~a, %eax" bool-tag))   ; add boolean type tag

(define (compile-primitive-call form si env)
  (case (primitive-op form)
    ; add1 - increment the first arg (fixnum) and return it
    ((add1)
     (compile-expr (primitive-op-arg1 form) si env)
     (emit "addl $~a, %eax" (immediate-rep 1)))

    ; sub1 - decrement the first arg (fixnum) and return it
    ((sub1)
     (compile-expr (primitive-op-arg1 form) si env)
     (emit "subl $~a, %eax" (immediate-rep 1)))

    ; integer? - check whether the first arg is an integer
    ((integer?)
     (compile-expr (primitive-op-arg1 form) si env)
     (emit "andl $~a, %eax" fixnum-mask)
     (emit-is-eax-equal-to 0))

    ; boolean? - check whether the first arg is a boolean
    ((boolean?)
     (compile-expr (primitive-op-arg1 form) si env)
     (emit "andl $~a, %eax" bool-mask)
     (emit-is-eax-equal-to bool-tag))

    ; char? check whether the first arg is a character
    ((char?)
     (compile-expr (primitive-op-arg1 form) si env)
     (emit "andl $~a, %eax" char-mask)
     (emit-is-eax-equal-to char-tag))

    ; zero? check whether the first arg is the fixnum 0
    ((zero?)
     (compile-expr (primitive-op-arg1 form) si env)
     (emit-is-eax-equal-to 0))

    ; + - addition of two fixnums
    ((+)
     (compile-expr (primitive-op-arg1 form) si env)
     (emit "movl %eax, ~a(%esp)" si)
     (compile-expr (primitive-op-arg2 form) (- si wordsize) env)
     (emit "addl ~a(%esp), %eax" si))

    ; - - subtraction of two fixnums
    ((-)
     (compile-expr (primitive-op-arg2 form) si env)
     (emit "movl %eax, ~a(%esp)" si)
     (compile-expr (primitive-op-arg1 form) (- si wordsize) env)
     (emit "subl ~a(%esp), %eax" si))

    ; * - multiplication of two fixnums
    ((*)
     (compile-expr (primitive-op-arg1 form) si env)
     (emit "movl %eax, ~a(%esp)" si)
     (compile-expr (primitive-op-arg2 form) (- si wordsize) env)
     (emit "shrl $~a, %eax" fixnum-shift) ; make sure only one is shifted
     (emit "imull ~a(%esp), %eax" si))

    ; = - fixnum equality test
    ((=)
     (compile-expr (primitive-op-arg1 form) si env)
     (emit "movl %eax, ~a(%esp)" si)
     (compile-expr (primitive-op-arg2 form) (- si wordsize) env)
     (emit "cmpl %eax, ~a(%esp)" si)
     (emit "movl $0, %eax")
     (emit "sete %al")
     (emit "sall $~a, %eax" bool-shift)
     (emit "orl $~a, %eax" bool-tag))

    ; = - fixnum less-than test
    ((<)
     (compile-expr (primitive-op-arg1 form) si env)
     (emit "movl %eax, ~a(%esp)" si)
     (compile-expr (primitive-op-arg2 form) (- si wordsize) env)
     (emit "cmpl %eax, ~a(%esp)" si)
     (emit "movl $0, %eax")
     (emit "setl %al")
     (emit "sall $~a, %eax" bool-shift)
     (emit "orl $~a, %eax" bool-tag))

    ; char=? - character equality test
    ((char=?)
     (compile-expr (primitive-op-arg1 form) si env)
     (emit "shrl $~a, %eax" char-shift) ; remove tag bits
     (emit "movl %eax, ~a(%esp)" si)
     (compile-expr (primitive-op-arg2 form) (- si wordsize) env)
     (emit "shrl $~a, %eax" char-shift) ; remove tag bits
     (emit "cmpl %eax, ~a(%esp)" si)
     (emit "movl $0, %eax")
     (emit "sete %al")
     (emit "sall $~a, %eax" bool-shift)
     (emit "orl $~a, %eax" bool-tag))

    ; cons - build a pair from the two arguments
    ((cons)
     ; evaluate x and y
     (compile-expr (primitive-op-arg1 form) si env)
     (emit "movl %eax, ~a(%esp)" si)
     (compile-expr (primitive-op-arg2 form) (- si wordsize) env)
     ; store cdr
     (emit "movl %eax, ~a(%esi)" wordsize)
     ; store car
     (emit "movl ~a(%esp), %eax" si)
     (emit "movl %eax, 0(%esi)")
     ; save pointer and tag it, then increment heap ptr
     (emit "movl %esi, %eax")
     (emit "orl $~a, %eax" pair-tag)
     (emit "addl $~a, %esi" (* 2 wordsize)))

    ; car - get the car of a pair
    ((car)
     (compile-expr (primitive-op-arg1 form) si env)
     (emit "movl ~a(%eax), %eax" (- pair-tag))) ; 0(%eax) plus tag offset

    ; cdr - get the cdr of a pair
    ((cdr)
     (compile-expr (primitive-op-arg1 form) si env)
     (emit "movl ~a(%eax), %eax" (- 4 pair-tag))) ; 4(%eax) plus tag offset

    ; null? - check whether an object is null
    ((null?)
     (compile-expr (primitive-op-arg1 form) si env)
     (emit-is-eax-equal-to pair-tag))

    ; (set-car! pair obj) - mutate the car of a pair
    ((set-car!)
     (compile-expr (primitive-op-arg1 form) si env) ; pair
     (emit "movl %eax, ~a(%esp)" si)
     (compile-expr (primitive-op-arg2 form) (- si wordsize) env) ; obj
     (emit "movl ~a(%esp), %edx" si)
     (emit "movl %eax, ~a(%edx)" (- pair-tag)))

    ; (set-car! pair obj) - mutate the car of a pair
    ((set-cdr!)
     (compile-expr (primitive-op-arg1 form) si env) ; pair
     (emit "movl %eax, ~a(%esp)" si)
     (compile-expr (primitive-op-arg2 form) (- si wordsize) env) ; obj
     (emit "movl ~a(%esp), %edx" si)
     (emit "movl %eax, ~a(%edx)" (- 4 pair-tag)))

    ; string? - return whether an object is a string
    ((string?)
     (compile-expr (primitive-op-arg1 form) si env)
     (emit "andl $~a, %eax" ptr-mask)
     (emit-is-eax-equal-to str-tag))

    ; (make-string k)   - construct a string of length k
    ; (make-string k c) - construct a string of length k filled with c
    ((make-string)
     (compile-expr (primitive-op-arg1 form) si env) ; deal with the length
     (emit "shr $~a, %eax" fixnum-shift) ; remove fixnum tag
     (emit "movl %eax, 0(%esi)") ; store length into new structure
     (emit "movl %eax, %ecx") ; save actual length while we tag the pointer
     (emit "movl %esi, %eax") ; tag the result ptr
     (emit "orl $~a, %eax" str-tag)
     (emit "addl $11, %ecx") ; pad size to the next multiple of 8 bytes
     (emit "andl $-8, %ecx")
     (emit "addl %ecx, %esi") ; advance allocation pointer
     (if (null? (cdddr form))
         '() ; no extra work for one-arg form
         (begin ; emit initialization code
           ; what should we fill with?
           (emit "movl %eax, ~a(%esp)" si) ; save the pointer
           (compile-expr (primitive-op-arg2 form) (- si wordsize) env)
           (emit "shrl $~a, %eax" char-shift) ; grab character into AL

           ; perform fill - rep stosb takes from AL, so we don't need to mask
           ; out any higher bits
           (emit "movl ~a(%esp), %edi" si) ; target pointer
           (emit "subl $~a, %edi" str-tag)
           (emit "movl 0(%edi), %ecx") ; load buffer length
           (emit "addl $~a, %edi" wordsize) ; skip over length word
           (emit "rep stosb") ; fill with byte

           (emit "movl ~a(%esp), %eax" si)))) ; recover final value

    ; string-length - get the length in chars of a string
    ((string-length)
     (compile-expr (primitive-op-arg1 form) si env)
     (emit "movl ~a(%eax), %eax" (- str-tag)) ; get length
     (emit "sall $~a, %eax" fixnum-shift)) ; convert to fixnum

    ; (string-ref str k) - get the kth character of str
    ((string-ref)
     (compile-expr (primitive-op-arg1 form) si env) ; get str
     (emit "subl $~a, %eax" str-tag) ; untag pointer
     (emit "addl $~a, %eax" wordsize) ; skip length word
     (emit "movl %eax, ~a(%esp)" si)
     (compile-expr (primitive-op-arg2 form) (- si wordsize) env)
     (emit "shrl $~a, %eax" fixnum-shift) ; get index
     (emit "addl ~a(%esp), %eax" si) ; offset pointer
     (emit "movzbl (%eax), %eax")
     (emit "shl $~a, %eax" char-shift) ; tag as char
     (emit "orl $~a, %eax" char-tag))

    ; (string-set! str k c) - set the kth character of str to c
    ((string-set!)
     (compile-expr (primitive-op-arg1 form) si env) ; get str
     (emit "subl $~a, %eax" str-tag) ; untag pointer
     (emit "addl $~a, %eax" wordsize) ; skip length word
     (emit "movl %eax, ~a(%esp)" si)
     (compile-expr (primitive-op-arg2 form) (- si wordsize) env)
     (emit "shrl $~a, %eax" fixnum-shift) ; get index
     (emit "addl %eax, ~a(%esp)" si) ; offset pointer
     (compile-expr (primitive-op-arg3 form) (- si wordsize) env) ; value
     (emit "shrl $~a, %eax" char-shift) ; get character value
     (emit "movl ~a(%esp), %ecx" si) ; store
     (emit "movb %al, (%ecx)")
     (emit "xorl %eax, %eax")) ; return 0

    ; vector? - return whether an object is a vector
    ((vector?)
     (compile-expr (primitive-op-arg1 form) si env)
     (emit "andl $~a, %eax" ptr-mask)
     (emit-is-eax-equal-to vec-tag))

    ; (make-vector k)   - construct a vector of length k
    ; (make-vector k o) - construct a vector of length k with cells set to o
    ((make-vector)
     (compile-expr (primitive-op-arg1 form) si env) ; deal with the length
     (emit "shr $~a, %eax" fixnum-shift) ; remove fixnum tag
     (emit "movl %eax, 0(%esi)") ; store length into new structure
     (emit "movl %eax, %ecx") ; save actual length
     (emit "sall $2, %ecx") ; multiply by 4 to get bytes
     (emit "movl %esi, %eax") ; tag the result ptr
     (emit "orl $~a, %eax" vec-tag)
     (emit "addl $11, %ecx") ; pad size to the next multiple of 8 bytes
     (emit "andl $-8, %ecx")
     (emit "addl %ecx, %esi") ; advance allocation pointer
     (if (null? (cdddr form))
         '() ; no extra work for one-arg form
         (begin ; emit initialization code
           ; what should we fill with?
           (emit "movl %eax, ~a(%esp)" si) ; save the pointer
           (compile-expr (primitive-op-arg2 form) (- si wordsize) env)

           ; load target pointer and count
           (emit "movl ~a(%esp), %edi" si) ; ptr
           (emit "subl $~a, %edi" vec-tag)
           (emit "movl 0(%edi), %ecx") ; count
           (emit "addl $~a, %edi" wordsize) ; don't write to length

           ; initialize items and return
           (emit "rep stosl") ; fill with byte
           (emit "movl ~a(%esp), %eax" si)))) ; recover final value

    ; vector-length - get the length in chars of a vector
    ((vector-length)
     (compile-expr (primitive-op-arg1 form) si env)
     (emit "movl ~a(%eax), %eax" (- vec-tag)) ; get length
     (emit "sall $~a, %eax" fixnum-shift)) ; convert to fixnum

    ; (vector-ref vec k) - get the kth element of vec
    ((vector-ref)
     (compile-expr (primitive-op-arg1 form) si env) ; get vec
     (emit "subl $~a, %eax" vec-tag) ; untag pointer
     (emit "addl $~a, %eax" wordsize) ; skip length word
     (emit "movl %eax, ~a(%esp)" si)
     (compile-expr (primitive-op-arg2 form) (- si wordsize) env)
     (emit "shrl $~a, %eax" (- fixnum-shift 2)) ; fixnum->idx (*4)
     (emit "addl ~a(%esp), %eax" si) ; offset pointer
     (emit "movl (%eax), %eax")) ; load and return

    ; (vector-set! vec k v) - set the kth element of vec to v
    ((vector-set!)
     (compile-expr (primitive-op-arg1 form) si env) ; get vec
     (emit "subl $~a, %eax" vec-tag) ; untag pointer
     (emit "addl $~a, %eax" wordsize) ; skip length word
     (emit "movl %eax, ~a(%esp)" si)
     (compile-expr (primitive-op-arg2 form) (- si wordsize) env)
     (emit "shrl $~a, %eax" (- fixnum-shift 2)) ; fixnum->idx (*4)
     (emit "addl %eax, ~a(%esp)" si) ; offset pointer
     (compile-expr (primitive-op-arg3 form) (- si wordsize) env) ; value
     (emit "movl ~a(%esp), %ecx" si) ; store
     (emit "movl %eax, (%ecx)")
     (emit "xorl %eax, %eax")) ; return 0
    ))

(define (let? x) (eq? (car x) 'let))

; build a list of numbers (a ... b-1)
(define (range a b)
  (if (= a b) '()
      (cons a (range (+ 1 a) b))))

(define (compile-let bindings body si env)
  (let* ((stack-offsets (map (lambda (x) (- si (* x wordsize)))
                             (range 0 (length bindings))))
         (names (map car bindings))
         (exprs (map cadr bindings))
         (inner-si (- si (* (length bindings) wordsize)))
         (inner-env (append (map cons names stack-offsets) env)))
    ; evaluate exprs and assign them to stack locations
    (for-each (lambda (expr offset)
                (compile-expr expr inner-si env)
                (emit "movl %eax, ~a(%esp)" offset))
              exprs stack-offsets)

    ; evaluate all body forms - this will leave the last one's output in %eax
    (for-each (lambda (form) (compile-expr form inner-si inner-env)) body)
    ))

(define (variable? x) (symbol? x))

(define (compile-var-load v si env)
  (emit "movl ~a(%esp), %eax" (cdr (assoc v env))))

(define current-label 0)
(define (unique-label)
  (let ((l current-label)) (set! current-label (+ 1 l))
    (format #f "l_~a" l)))

(define (if? x) (eq? (car x) 'if))

(define (compile-if test t-body f-body si env)
  (let ((false-label (unique-label))
        (end-label (unique-label)))
    (compile-expr test si env)
    (emit "cmpl $~a, %eax" (immediate-rep #f))
    (emit "je ~a" false-label)
    (compile-expr t-body si env)
    (emit "jmp ~a" end-label)
    (emit "~a:" false-label)
    (compile-expr f-body si env)
    (emit "~a:" end-label)))

(define (compile-expr e si env)
  (cond
    ((immediate? e) (emit "movl $~a, %eax" (immediate-rep e)))
    ((variable? e) (compile-var-load e si env))
    ((let? e) (compile-let (cadr e) (cddr e) si env))
    ((if? e) (compile-if (cadr e) (caddr e)
                         (if (null? (cdddr e)) #f (cadddr e))
                         si env))
    ((primitive-call? e) (compile-primitive-call e si env))
    ))

(define (compile-program program)
  (set! current-label 0)
  (emit ".text")
  (emit ".p2align 4,,15")
  (emit ".globl scheme_entry")
  (emit "scheme_entry:")
  
  ; handle incoming call from C
  (emit "push %esi")
  (emit "push %edi")
  (emit "push %edx")
  (emit "movl 16(%esp), %esi")
  
  ; our code goes here
  (compile-expr program (- wordsize) '())
  
  ; restore state for return to C
  (emit "pop %edx")
  (emit "pop %edi")
  (emit "pop %esi")
  (emit "ret")
  )

(define (compile-to-binary program)
  (begin
    (with-output-to-file "/tmp/compiled.s"
      (lambda () (compile-program program)))
    (system "gcc -fomit-frame-pointer -m32 /tmp/compiled.s rts.c")))

(define (compile-and-run program)
  (begin (compile-to-binary program)
         (system "./a.out")))
